/*------------------------------------------------*/
/* 10assimp/assimp/main.c                         */
#define _CRT_SECURE_NO_WARNINGS
#define GLEW_STATIC
#include <GL/glew.h>
#include <stdio.h>
#include <GLFW/glfw3.h>
#include <stdbool.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include "linmath-extended.h"
#include "vertex-data.h"
#include "shader.h"
#include "camera.h"
#include "events.h"
#include "stb-image.h"
#include "load-cubemap.h"
#include "model.h"

void initVariables(Shader *program);
void initBuffers(Shader *program);
void uninitBuffers(void);
void initCubes(void);
void ticking(Shader * program, GLFWwindow* window);
void animate(void);
void draw(Shader *program, GLFWwindow* window, GLuint *VAO);
static void drawCube(GLfloat ratio, GLfloat scale, GLfloat x, GLfloat y, GLfloat z, vec3 * position);
static void drawModel(GLfloat ratio, GLfloat scale, GLfloat x, GLfloat y, GLfloat z, Model *model);

GLuint VAO[1], BO[2], TEX[1];
struct {
	GLuint uPMatrixLoc;
	GLuint uMVMatrixLoc;
	GLuint texture1;
} variables;

GLfloat deltaTime = 0.0f;
GLfloat lastFrame = 0.0f;

#define CUBES_COUNT 500
GLfloat cubes[CUBES_COUNT][4];

Model model;

int main(void)
{
	GLFWwindow* window;
	/* Initialize the library */
	if (!glfwInit()) {
		return -1;
	}
	glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
	glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
	glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);

	/* Create a windowed mode window and its OpenGL context */
	GLFWmonitor* monitor = glfwGetPrimaryMonitor(); 	
	const GLFWvidmode* mode = glfwGetVideoMode(monitor);
	glfwWindowHint(GLFW_RED_BITS, mode->redBits);
	glfwWindowHint(GLFW_GREEN_BITS, mode->greenBits);
	glfwWindowHint(GLFW_BLUE_BITS, mode->blueBits);
	glfwWindowHint(GLFW_REFRESH_RATE, mode->refreshRate);
	window = glfwCreateWindow(mode->width, mode->height, "10 - assimp", monitor, NULL);
	if (!window)
	{
		printf("Nelze inicializovat kontext\n");
		glfwTerminate();
		return -1;
	}
	/* Make the window's context current */
	glfwMakeContextCurrent(window);
	printf("%s\n", glGetString(GL_VERSION));

	/* Load modern OpenGL functions */
	glewExperimental = GL_TRUE;
	glewInit();

	/* events */
	glfwSetKeyCallback(window, key_callback);
	glfwSetScrollCallback(window, scroll_callback);
	glfwSetCursorPosCallback(window, cursor_position_callback);
	
	glfwSetInputMode(window, GLFW_CURSOR, GLFW_CURSOR_DISABLED);

	//glfwSwapInterval(1);
	camera = camera3f(0.0f, 0.0f, 0.0f);
	camera.MovementSpeed = 50.f;
	// http://www.learnopengl.com/#!Model-Loading/Model
	model = constructModel("models/nanosuit/nanosuit.obj");
	// http://tf3dm.com/3d-model/superman-39631.html
	// model = constructModel("models/superman/SuperMan.obj");
	Shader program = shader("cubemap.vs", "cubemap.fs");
	if (!program.Program || !model.success) {
		return -1;
	}

	initVariables(&program);
	initBuffers(&program);
	initCubes();

	// setup GL state
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	ticking(&program, window);

	uninitBuffers();

	glfwDestroyWindow(window);
	glfwTerminate();
	model.destroy(&model);
	return 0;
}

void initVariables(Shader *program)
{
	variables.uMVMatrixLoc = program->getUniformLocation(program, "uMVMatrix");
	variables.uPMatrixLoc = program->getUniformLocation(program, "uPMatrix");
	variables.texture1 = program->getUniformLocation(program, "texture1");
}

void initBuffers(Shader *program)
{
	glGenVertexArrays(sizeof(VAO) / sizeof(VAO[0]), VAO);
	glGenBuffers(sizeof(BO) / sizeof(BO[0]), BO);
	glGenTextures(sizeof(TEX) / sizeof(TEX[0]), TEX);

	const char * const faces[] = {
		"cubemaps/miramar_large/right.jpg",
		"cubemaps/miramar_large/left.jpg",
		"cubemaps/miramar_large/top.jpg",
		"cubemaps/miramar_large/bottom.jpg",
		"cubemaps/miramar_large/back.jpg",
		"cubemaps/miramar_large/front.jpg",
	};

	glActiveTexture(GL_TEXTURE0);
	load_cubemap(TEX[0], faces);

	glBindVertexArray(VAO[0]);
	{
		//vertices
		glBindBuffer(GL_ARRAY_BUFFER, BO[0]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(boxVertices), boxVertices, GL_STATIC_DRAW);
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
		glEnableVertexAttribArray(0);
		//textures
		glBindBuffer(GL_ARRAY_BUFFER, BO[1]);
		glBufferData(GL_ARRAY_BUFFER, sizeof(boxTexCoord), boxTexCoord, GL_STATIC_DRAW);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 3 * sizeof(GLfloat), (GLvoid*)0);
		glEnableVertexAttribArray(1);
	}
	glBindVertexArray(0); // Unbind VAO
}

void uninitBuffers(void)
{
	glDeleteVertexArrays(sizeof(VAO) / sizeof(VAO[0]), VAO);
	glDeleteBuffers(sizeof(BO) / sizeof(BO[0]), BO);
}

void initCubes(void)
{
	int i;
	srand(time(NULL));

	for (i = 0; i < CUBES_COUNT; i++) {
		cubes[i][0] = 0.01f; // scale
		cubes[i][1] = 2000.0 * rand()/ RAND_MAX - 1000.0; //x
		cubes[i][2] = -10.0f; // +500.0 * rand() / RAND_MAX; // y
		cubes[i][3] = 2000.0 * rand()/ RAND_MAX - 1000.0; //z
	}
};

void ticking(Shader * program, GLFWwindow* window) {
	/* Loop until the user closes the window */
	while (!glfwWindowShouldClose(window))
	{
		animate();
		draw(program, window, VAO);
		glfwSwapBuffers(window);
		glfwPollEvents();
	}
}

void animate(void) {
	// Set frame time
	GLfloat currentFrame = (GLfloat) glfwGetTime();
	deltaTime = currentFrame - lastFrame;
	lastFrame = currentFrame;
	// Camera controls
	if (keys[GLFW_KEY_W])
		camera.ProcessKeyboard(&camera, FORWARD, deltaTime);
	if (keys[GLFW_KEY_S])
		camera.ProcessKeyboard(&camera, BACKWARD, deltaTime);
	if (keys[GLFW_KEY_Q])
		camera.ProcessKeyboard(&camera, LEFT, deltaTime);
	if (keys[GLFW_KEY_E])
		camera.ProcessKeyboard(&camera, RIGHT, deltaTime);
	if (keys[GLFW_KEY_A])
		camera.ProcessKeyboard(&camera, TURN_LEFT, deltaTime);
	if (keys[GLFW_KEY_D])
		camera.ProcessKeyboard(&camera, TURN_RIGHT, deltaTime);
}


void draw(Shader *program, GLFWwindow* window, GLuint *VAO)
{
	int width, height;
	glfwGetFramebufferSize(window, &width, &height);
	glViewport(0, 0, width, height);
	GLfloat ratio = ((GLfloat)width) / height;

	glClearColor(1.f, 1.f, 1.f, 1.f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	program->Use(program);
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_CUBE_MAP, TEX[0]);
	vec3 center = { 0.0, 0.0, 0.0 };
	drawCube(ratio, 1.0, 0.0, 0.0, 0.0, &center);
	glClear(GL_DEPTH_BUFFER_BIT);

	glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
	int i;
	for (i = 0; i < sizeof(cubes) / sizeof(cubes[0]); i++) {
		drawCube(ratio, cubes[i][0], cubes[i][1], cubes[i][2], cubes[i][3], NULL);
	}

	drawModel(ratio, 10.0, 0.0, -20.0, -100.0, &model);
	drawModel(ratio, 10.0, 100.0, -20.0, -100.0, &model);
	drawModel(ratio, 10.0, -100.0, -20.0, -100.0, &model);
}

static void drawCube(GLfloat ratio, GLfloat scale, GLfloat x, GLfloat y, GLfloat z, vec3 * cameraPosition)
{
	mat4x4 uMVMatrix, uPMatrix, view;
	mat4x4_identity(uMVMatrix);
	mat4x4_identity(uPMatrix);
	mat4x4_perspective(uPMatrix, degToRad(camera.Zoom), ratio, 1.0, 20000.0);
	mat4x4_translate(uMVMatrix, x, y, z);
	mat4x4_scale_aniso(uMVMatrix, uMVMatrix, scale, scale, scale);

	// Create camera transformation
	camera.GetViewMatrix(&camera, view, cameraPosition);
	mat4x4_mul(uMVMatrix, view, uMVMatrix);
	//mat4x4_print(uPMatrix);
	//mat4x4_print(uMVMatrix);
	glUniformMatrix4fv(variables.uMVMatrixLoc, 1, GL_FALSE, (const GLfloat *)uMVMatrix);
	glUniformMatrix4fv(variables.uPMatrixLoc, 1, GL_FALSE, (const GLfloat *)uPMatrix);

	// vykresli
	glBindVertexArray(VAO[0]);
	glDrawArrays(GL_TRIANGLES, 0, sizeof(boxVertices) / sizeof(boxVertices[0]) / 3);
	glBindVertexArray(0);
}

static void drawModel(GLfloat ratio, GLfloat scale, GLfloat x, GLfloat y, GLfloat z, Model *model)
{
	mat4x4 uMVMatrix, uPMatrix, view;
	mat4x4_identity(uMVMatrix);
	mat4x4_identity(uPMatrix);
	mat4x4_perspective(uPMatrix, degToRad(camera.Zoom), ratio, 1.0, 20000.0);
	mat4x4_translate(uMVMatrix, x, y, z);
	mat4x4_scale_aniso(uMVMatrix, uMVMatrix, scale, scale, scale);

	// Create camera transformation
	camera.GetViewMatrix(&camera, view, NULL);
	mat4x4_mul(uMVMatrix, view, uMVMatrix);

	model->draw(model, uMVMatrix, view, uPMatrix);
}
/*------------------------------------------------*/
